@@ -1,7 +1,7 @@
 /*
  * uicode_tool - Manipulates Intel(R) IA32/x86_64 processor microcode bundles
  *
- * Copyright (c) 2010-2012 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
+ * Copyright (c) 2010-2013 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
  *               2000 Simon Trimmer, Tigran Aivazian
  *
  * Some code copied from microcode.ctl v1.17.
@@ -27,6 +27,7 @@
 #include <assert.h>
 #include <argp.h>
 #include <dirent.h>
+#include <time.h>
 
 #include "intel_microcode.h"
 #include "iucode_tool_config.h"
@@ -39,6 +40,11 @@
 #define likely(x)       __builtin_expect(!!(x), 1)
 #define unlikely(x)     __builtin_expect(!!(x), 0)
 
+/* cpio archive constants */
+#define LINUX_CPIO_UCODE_NAME "kernel/x86/microcode/GenuineIntel.bin"
+#define LINUX_CPIO_FILE_MODE 0100644
+#define LINUX_CPIO_DIR_MODE  0040755
+
 /* Command-line processing and UI */
 
 enum {
@@ -46,20 +52,22 @@ enum {
 	EXIT_SWFAILURE = 2,	/* Exit due to software failure (ENOMEM, etc) */
 };
 
-enum { /* actions */
+static enum { /* actions */
 	IUCODE_DO_UPLOADUC = 0x0001,
 	IUCODE_DO_WRITEUC  = 0x0002,
 	IUCODE_DO_WRITEFW  = 0x0004,
 	IUCODE_DO_SELPROC  = 0x0008,
 	IUCODE_DO_LOADFILE = 0x0010,
 	IUCODE_DO_WRITEFWN = 0x0020,
+	IUCODE_DO_WRITEFWE = 0x0040,
 
 	IUCODE_F_UCSELECT  = 0x1000,
 
 	IUCODE_DOMASK_NEEDSUC = IUCODE_DO_UPLOADUC
 				| IUCODE_DO_WRITEUC
 				| IUCODE_DO_WRITEFW
-				| IUCODE_DO_WRITEFWN,
+				| IUCODE_DO_WRITEFWN
+				| IUCODE_DO_WRITEFWE,
 } command_line_actions;
 
 static char *progname;
@@ -72,6 +80,7 @@ static int list_sel_microcodes = 0;
 static int unlink_files = 0;
 static char *upload_microcode = NULL;
 static char *write_microcode = NULL;
+static char *write_early_firmware = NULL;
 static char *write_firmware = NULL;
 static char *write_named = NULL;
 
@@ -102,10 +111,18 @@ struct microcode_bundle {
 static struct microcode_bundle *microcode_bundles = NULL;
 static unsigned long int last_bundle_id = 0;
 
+/* intel_uclist_entry signature flag constants */
+enum {
+    INTEL_UCLE_EXTSIG = 0x0001,	/* Sig came from extended table */
+    INTEL_UCLE_SELID  = 0x0010,	/* Sig is candidate for loose datefilter */
+    INTEL_UCLE_SELECT = 0x0020,	/* Sig selected */
+};
+
 /* signature single-linked list */
 struct intel_uclist_entry {
 	unsigned long int id;	/* microcode id within group */
 	unsigned long int gid;	/* microcode group id */
+	uint32_t flags;		/* INTEL_UCLE_ flags */
 	uint32_t cpuid;		/* sorting key */
 	uint32_t pf_mask;
 	int32_t  uc_rev;	/* duplicated from header */
@@ -113,6 +130,7 @@ struct intel_uclist_entry {
 	uint32_t uc_size;
 	struct intel_uclist_entry *next;
 };
+static struct intel_uclist_entry *all_microcodes = NULL;
 static struct intel_uclist_entry *microcodes = NULL;
 
 /* UI helpers */
@@ -120,6 +138,7 @@ static struct intel_uclist_entry *microcodes = NULL;
 struct microcode_interator_data {
 	const struct microcode_bundle *current_bundle;
 	unsigned long int total_signature_count;
+	unsigned long int total_unique_sig_count;
 	unsigned long int total_entry_count;
 	unsigned long int current_uc;
 	char current_uc_id_str[UCODE_ID_MAX_LEN];
@@ -138,6 +157,7 @@ static int filter_list_allow = 1;
 
 static uint32_t datefilter_max = 0xffffffffU;  /* select dates before this */
 static uint32_t datefilter_min = 0;	       /* select dates after this */
+static int datefilter_loose = 0;
 
 static int inline filter_list_active(void)
 {
@@ -181,7 +201,7 @@ static void str_append_ucode_id(unsigned long int id,
  * If p != NULL, frees all members of the struct intel_uclist_entry
  * linked list.
  */
-void free_uclist(struct intel_uclist_entry *p)
+static void free_uclist(struct intel_uclist_entry *p)
 {
 	struct intel_uclist_entry *e;
 
@@ -206,7 +226,7 @@ static void __uclist_remove_shadowed(const unsigned long int gid,
 		struct intel_uclist_entry * const e = *pnext;
 
 		if (((pf_mask & e->pf_mask) == e->pf_mask) &&
-		    (uc_rev >= e->uc_rev || (downgrade && gid != e->gid))) {
+		    (uc_rev >= e->uc_rev || (downgrade && gid > e->gid))) {
 			/* entry is a subset of the one we just added */
 			/* delete entry */
 			*pnext = e->next;
@@ -218,7 +238,94 @@ static void __uclist_remove_shadowed(const unsigned long int gid,
 }
 
 /**
- * uclist_add_signature() - adds signature to ucode sig list
+ * uclist_add_signature() - add signature to ucode sig list
+ *
+ * @id:		microcode id
+ * @gid:	microcode group id
+ * @flags:	INTEL_UCLE_* flags
+ * @cpuid:	signature of the signature entry being added
+ * @pf_mask:	pf_mask of the signature entry being added
+ * @uc_rev:	revision of the microcode entry being added
+ * @uc:		microcode data (including headers)
+ * @uc_size:	microcode data size (total, including headers)
+ * @strict:	validate microcode opaque data against duplicates
+ * @uclist:	pointer to the first entry on the list
+ *
+ * Adds a microcode signature entry to the list.  Does not filter
+ * out duplicates, nor superseed entries.
+ *
+ * If @strict is true, it will compare the opaque data with one
+ * of the previously inserted duplicates (and return EBADF if it
+ * doesn't match).
+ *
+ * returns: ENOMEM: could not allocate memory,
+ *          EINVAL: bad @cpuid, @uc pointer or @uclist pointer
+ *          EEXIST: entry already in list (will be inserted)
+ *          EBADF : entry already in list, with different opaque data
+ *          0:      entry not yet in list (will be inserted)
+ */
+static int uclist_add_signature(const unsigned long int id,
+			const unsigned long int gid,
+			const uint32_t flags,
+			const uint32_t cpuid,
+			const uint32_t pf_mask,
+			const int32_t uc_rev,
+			const void * const uc,
+			const uint32_t uc_size,
+			const int strict,
+			struct intel_uclist_entry ** const uclist)
+{
+	struct intel_uclist_entry **pnext, *n;
+	int res = 0;
+
+	if (unlikely(!cpuid || !uc || !uclist))
+		return EINVAL;
+
+	pnext = uclist;
+	n = NULL;
+
+	/* search linked list for first insertion point */
+	while (likely(*pnext && (*pnext)->cpuid < cpuid))
+		pnext = &((*pnext)->next);
+
+	/* look for duplicates */
+	while (likely(*pnext) && (*pnext)->cpuid == cpuid) {
+		if ((*pnext)->pf_mask == pf_mask && (*pnext)->uc_rev == uc_rev) {
+			res = EEXIST;
+			n = *pnext;
+		}
+		pnext = &((*pnext)->next);
+	}
+
+	if (strict && res == EEXIST) {
+	    	int dup = intel_ucode_compare(uc, n->uc);
+		if (dup == EINVAL || dup == EBADF)
+			return dup;
+	}
+
+	n = malloc(sizeof(struct intel_uclist_entry));
+	if (unlikely(!n))
+		return ENOMEM;
+
+	memset(n, 0, sizeof(struct intel_uclist_entry));
+	n->id = id;
+	n->gid = gid;
+	n->cpuid = cpuid;
+	n->pf_mask = pf_mask;
+	n->uc_rev = uc_rev;
+	n->uc = uc;
+	n->uc_size = uc_size;
+	n->flags = flags;
+
+	/* prepend */
+	n->next = *pnext;
+	*pnext = n;
+
+	return res;
+}
+
+/**
+ * uclist_merge_signature() - merge signature into ucode sig list
  * @id:		microcode id
  * @gid:	microcode group id
  * @cpuid:	signature of the signature entry being added
@@ -233,7 +340,7 @@ static void __uclist_remove_shadowed(const unsigned long int gid,
  * is zero, it will replace any suitable entry that has a lesser
  * revision.  When @downgrade is non-zero, it will replace any
  * suitable entry that has a lesser revision or which is from
- * a different microcode group.
+ * a smaller (earlier) microcode group.
  *
  * "suitable" means microcodes with the same cpuid and with a
  * pf_mask that is either equal or a strict subset of the one from
@@ -248,7 +355,7 @@ static void __uclist_remove_shadowed(const unsigned long int gid,
  *                  inserted if @downgrade is non-zero)
  *          EINVAL: bad @cpuid, @uc pointer or @uclist pointer
  */
-int uclist_add_signature(const unsigned long int id,
+static int uclist_merge_signature(const unsigned long int id,
 			const unsigned long int gid,
 			const uint32_t cpuid,
 			const uint32_t pf_mask,
@@ -277,7 +384,7 @@ int uclist_add_signature(const unsigned long int id,
 		if (likely(pf_mask && (e->pf_mask & pf_mask) == 0)) {
 			/* disjoint sets, continue search */
 			pnext = &(e->next);
-		} else if (!downgrade || gid == e->gid) { /* do not downgrade */
+		} else if (!downgrade || gid <= e->gid) { /* do not downgrade */
 			if ((e->pf_mask & pf_mask) == pf_mask) {
 				/* new set same or contained in the old set */
 				if (e->uc_rev >= uc_rev) {
@@ -694,7 +801,7 @@ err_out:
 	return err;
 }
 
-static int load_intel_microcode(const char *path,
+static int load_intel_microcode(const char * path,
 				const intel_ucode_file_type_t baseftype)
 {
 	int fd = -1;
@@ -792,6 +899,9 @@ static int load_intel_microcode(const char *path,
 			continue;
 		}
 
+		if (S_ISREG(st.st_mode) && st.st_size == 0)
+			continue;
+
 		ftype = baseftype;
 		if (ftype == INTEL_UC_FT_UNKNOWN && S_ISREG(st.st_mode)) {
 			char *p;
@@ -828,18 +938,18 @@ static int load_intel_microcode(const char *path,
 
 /* Microcode write (binary format) and kernel upload */
 
-static ssize_t __write_intel_microcode_entry(const int fd,
-						const void * const uc,
-						const uint32_t uc_size)
+static ssize_t __write_data(const int fd,
+			    const void * const data,
+			    const uint32_t size)
 {
 	const void *p;
 	ssize_t len, s;
 
-	len = uc_size;
+	len = size;
 	if (len < 0)
 		return -EINVAL;
 
-	p = uc;
+	p = data;
 	while (len > 0) {
 		s = write(fd, p, len);
 		if (s == -1) {
@@ -851,7 +961,7 @@ static ssize_t __write_intel_microcode_entry(const int fd,
 		}
 	}
 
-	return uc_size;
+	return size;
 }
 
 static void log_microcode_action(const char * const action,
@@ -909,9 +1019,8 @@ static int upload_intel_microcodes(const char * const devname,
 			log_microcode_action("uploading",
 					     devname, uc_write_list);
 
-		rc = __write_intel_microcode_entry(fd,
-					uc_write_list->uc,
-					uc_write_list->uc_size);
+		rc = __write_data(fd, uc_write_list->uc,
+				  uc_write_list->uc_size);
 		if (rc < 0) {
 			err = -rc;
 			print_err("%s: write error: %s",
@@ -937,21 +1046,149 @@ static int upload_intel_microcodes(const char * const devname,
 	return err;
 }
 
+#define CPIO_STATIC_HDRSIZE (6 + 13*8)
+#define CPIO_MAX_HDRSIZE (CPIO_STATIC_HDRSIZE + sizeof(LINUX_CPIO_UCODE_NAME))
+
+/* if size is zero, assume it is a directory */
+static int __write_cpio_hdrentry(int fd, const char * const name,
+				 const size_t size)
+{
+	static int ino = 100;
+
+	char buf[CPIO_MAX_HDRSIZE + 4];
+	int nsize;
+	int bufsize;
+
+	assert(name);
+	nsize = strlen(name) + 1;
+
+	ino++;
+
+	/* pad to DWORD. Caller guarantees last write was also padded */
+	bufsize = CPIO_STATIC_HDRSIZE + nsize;
+	while (bufsize % 4)
+		bufsize++;
+
+	memset(buf, 0, sizeof(buf));
+	snprintf(buf, sizeof(buf),
+	  "070701" /* signature, new ASCII portable format, no CRC */
+	  "%08X%08X%08X%08X%08X%08lX%08zX%08X%08X%08X%08X%08X%08X%s",
+	  ino,							   /* inode */
+	  size ? LINUX_CPIO_FILE_MODE : LINUX_CPIO_DIR_MODE,	    /* mode */
+	  0, 0,							/* uid, gid */
+	  size ? 1 : 2, (long int) time(NULL), size,  /* nlink, mtime, size */
+	  3, 1, 0, 0, nsize, 0,                   /* devj, devm, nsize, CRC */
+	  name);                                               /* name, pad */
+
+	return __write_data(fd, buf, bufsize);
+}
+
+/**
+ * __write_cpio_header() - writes cpio header for early microcode
+ * @fd:			file to write to
+ * @size:		size of the file being written, for the header
+ *
+ * returns the number of bytes written if successful, or -errno()
+ * if a problem happens
+ *
+ * Write cpio header.  For Intel microcodes, the kernel ABI
+ * defines that they should be in a single file, named
+ * kernel/x86/microcode/GenuineIntel.bin.  We use the 070701
+ * format ("newc" format for GNU cpio).
+ */
+static int __write_cpio_header(int fd, size_t size)
+{
+	char fn[] = LINUX_CPIO_UCODE_NAME;
+	char *p1, *p2;
+	int s = 0;
+	int r;
+
+	p1 = fn;
+	do {
+		p2 = strchr(p1, '/');
+		if (p2) {
+			*p2 = 0;
+			r = __write_cpio_hdrentry(fd, fn, 0);
+			if (r < 0)
+				return r;
+			s += r;
+			*p2 = '/';
+			p1 = ++p2;
+		}
+	} while (p2);
+
+	r = __write_cpio_hdrentry(fd, fn, size);
+	if (r < 0)
+		return r;
+
+	s += r;
+	return s;
+}
+
+/**
+ * __write_cpio_trailer() - write a cpio EOF trailer and pad
+ * @fd:			file to write to
+ * @size:		file position/current size
+ *
+ * returns the number of bytes written if successful, or -errno()
+ * if a problem happens
+ *
+ * Write a cpio EOF trailer, padding with NULs to the nearest
+ * 16-byte boundary.  @size is used to calculate the amount of
+ * padding, instead of requiring a lseek().
+ */
+static int __write_cpio_trailer(int fd, size_t size)
+{
+	const char cpio_trailer[] = "070701"
+		/* inode, mode,  uid,   gid */
+		"00000000" "00000000" "00000000" "00000000" 
+		/* nlink, mtime, size,  devj */
+		"00000001" "00000000" "00000000" "00000000"
+		/* devm,  rdevj, rdevm, name size */
+		"00000000" "00000000" "00000000" "0000000B"
+		/* checksum, name */
+		"00000000" "TRAILER!!!"
+		/* 16 extra NULs for padding */
+		"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
+
+	size_t s = sizeof(cpio_trailer) - 16;
+	while ((size + s) % 16)
+		s++;
+
+	return __write_data(fd, cpio_trailer, s);
+}
+
+
 /**
  * write_intel_microcodes() - writes microcodes to bin file
  * @filename:		file to write to
+ * @ft:			file type
  * @uc_write_list:	uclist with the microcodes to write
  *
  * returns 0 if successful, or errno() if a problem happens
+ *
+ * @ft should be zero for binary format, 1 for linux early
+ * initramfs cpio format.
+ *
+ * binary format is the Intel-specified format for binary microcode,
+ * used by the Linux firmware loader and a few other operating
+ * systems.
+ *
+ * Linux early initramfs cpio format will write microcodes inside
+ * a cpio 070701 (New portable ASCII format without CRC) wrapper,
+ * suitable for early microcode loading for Linux kernels v3.9 and
+ * newer.  This file should be prepended to the compressed initramfs
+ * image.
  */
-static int write_intel_microcodes(const char * const filename,
+static int write_intel_microcodes(const char * const filename, int ft,
 				  struct intel_uclist_entry *uc_write_list)
 {
 	int fd;
 	int err = 0;
-	size_t total_written = 0;
+	size_t total_written;
 	unsigned int entries_written = 0;
 	ssize_t rc;
+	struct intel_uclist_entry *e;
 
 	assert(uc_write_list);
 	assert(filename);
@@ -978,24 +1215,43 @@ static int write_intel_microcodes(const char * const filename,
 		return err;
 	}
 
+	/*
+	 * precalculate raw file size. It is less annoying than
+	 * having to seek back to write it later in cpio mode
+	 */
+	e = uc_write_list;
+	total_written = 0;
+	while (e && e->uc) {
+		total_written += e->uc_size;
+		e = e->next;
+	}
+
+	if (ft == 1) {
+		rc = __write_cpio_header(fd, total_written);
+		if (rc < 0)
+			goto error;
+		total_written += rc;
+	}
+
 	while (uc_write_list && uc_write_list->uc) {
 		if (verbosity >= 3)
 			log_microcode_action("writing",
 					     filename, uc_write_list);
 
-		rc = __write_intel_microcode_entry(fd,
-					uc_write_list->uc,
-					uc_write_list->uc_size);
-		if (rc < 0) {
-			err = -rc;
-			print_err("%s: write error: %s",
-				  filename, strerror(err));
-			break;
-		} else {
-			total_written += rc;
-			uc_write_list = uc_write_list->next;
-			entries_written++;
-		}
+		rc = __write_data(fd, uc_write_list->uc,
+				  uc_write_list->uc_size);
+		if (unlikely(rc < 0))
+			goto error;
+
+		uc_write_list = uc_write_list->next;
+		entries_written++;
+	}
+
+	if (ft == 1) {
+		rc = __write_cpio_trailer(fd, total_written);
+		if (unlikely(rc < 0))
+			goto error;
+		total_written += rc;
 	}
 
 	if (close(fd)) {
@@ -1009,6 +1265,11 @@ static int write_intel_microcodes(const char * const filename,
 			  filename, entries_written, total_written);
 
 	return err;
+
+error:
+	err = -rc;
+	print_err("%s: write error: %s", filename, strerror(err));
+	return err;
 }
 
 /* Microcode filtering */
@@ -1147,13 +1408,44 @@ static int __uclist_add_print_errors(int status)
 		print_err("Cannot add index entry: out of memory");
 		return 1;
 	case EINVAL:
-		print_err("Internal error uclist_add_signature() returned EINVAL");
+		print_err("Internal error uclist_merge_signature() returned EINVAL");
 		return 1;
 	default:
 		return 1;
 	}
 }
 
+static int __datefilter_loose_inplaceinsert(struct intel_uclist_entry **p,
+					     struct intel_uclist_entry **uclist)
+{
+	const uint32_t cpuid = (*p)->cpuid;
+	const uint32_t pfm   = (*p)->pf_mask;
+	struct intel_uclist_entry *e;
+	int rc;
+
+	e = *uclist;
+	while (e && e->cpuid < cpuid)
+		e = e->next;
+	*uclist = e; /* speed up next search */
+
+	while (e && e->cpuid == cpuid) {
+		if (e->flags & INTEL_UCLE_SELID && e->pf_mask & pfm) {
+			e->flags &= ~INTEL_UCLE_SELID;
+			e->flags |= INTEL_UCLE_SELECT;
+			rc = __uclist_add_print_errors(
+			        uclist_merge_signature(e->id, e->gid,
+				        e->cpuid, e->pf_mask,
+					e->uc_rev, e->uc, e->uc_size,
+					allow_downgrade, p));
+			if (rc)
+				return rc;
+		}
+		e = e->next;
+	}
+
+	return 0;
+}
+
 static int __process_ucode_signature(void *userdata,
 				unsigned int const sig_count,
 				uint32_t const cpuid,
@@ -1166,7 +1458,8 @@ static int __process_ucode_signature(void *userdata,
 	struct microcode_interator_data * const ctx = userdata;
 	struct intel_ucode_metadata m;
 	intel_ucode_status_t s;
-	int add_status = ENOENT;
+	int add_status;
+	uint32_t uce_flags;
 
 	assert(ctx);
 
@@ -1179,19 +1472,10 @@ static int __process_ucode_signature(void *userdata,
 		return 1;
 	}
 
-	if (is_in_date_range(&m) &&
-	    is_selected(cpuid, pf_mask, uc_filter_list)) {
-		add_status = uclist_add_signature(ctx->current_uc,
-					ctx->current_bundle->id,
-					cpuid, pf_mask, m.revision,
-					uc, uc_size, allow_downgrade,
-					&microcodes);
-		if (add_status && add_status != EEXIST) {
-			print_err("Failed to select microcode entry %s: %s",
-				  ctx->current_uc_id_str, strerror(add_status));
-			return 1;
-		}
-	}
+	uce_flags = (sig_count)? INTEL_UCLE_EXTSIG : 0;
+	if (is_selected(cpuid, pf_mask, uc_filter_list))
+		uce_flags |= (is_in_date_range(&m)) ?
+				INTEL_UCLE_SELECT : INTEL_UCLE_SELID;
 
 	if (list_all_microcodes) {
 		if (!sig_count)
@@ -1208,16 +1492,42 @@ static int __process_ucode_signature(void *userdata,
 				m.revision);
 	}
 
+	add_status = uclist_add_signature(ctx->current_uc,
+			ctx->current_bundle->id, uce_flags,
+			cpuid, pf_mask, m.revision, uc, uc_size,
+			strict_checks, &all_microcodes);
+
 	switch (add_status) {
 	case 0:
-	case ENOENT:
+		ctx->total_unique_sig_count++;
+		break;
 	case EEXIST:
 		break;
+	case EBADF:
+		print_err("WARNING: Microcode %s has the same revision "
+			  "and signature as a previously loaded microcode, "
+			  "but different contents", ctx->current_uc_id_str);
+		return 1;
 	default:
-		return __uclist_add_print_errors(add_status);
+		print_err("Failed to add microcode entry %s: %s",
+			  ctx->current_uc_id_str, strerror(add_status));
+		return 1;
 	}
 
-	return 0;
+	if (uce_flags & INTEL_UCLE_SELECT) {
+		add_status = uclist_merge_signature(ctx->current_uc,
+					ctx->current_bundle->id,
+					cpuid, pf_mask, m.revision,
+					uc, uc_size, allow_downgrade,
+					&microcodes);
+		if (add_status && add_status != EEXIST) {
+			print_err("Failed to select microcode entry %s: %s",
+				  ctx->current_uc_id_str, strerror(add_status));
+			return 1;
+		}
+	}
+
+	return __uclist_add_print_errors(add_status);
 }
 
 static int __process_ucode_entry(void *userdata,
@@ -1253,42 +1563,6 @@ static int __process_ucode_entry(void *userdata,
 	return 0;
 }
 
-static int __show_ucode_signature_and_count(void *userdata,
-				unsigned int const sig_count,
-				uint32_t const cpuid,
-				uint32_t const pf_mask,
-				const void * const uc_data,
-				unsigned int const uc_data_size,
-				const void * const uc,
-				unsigned int const uc_size)
-{
-	unsigned long int * const counter = userdata;
-
-	assert(counter);
-
-	(*counter)++;
-
-	if (list_sel_microcodes) {
-		struct intel_ucode_metadata m;
-
-		intel_ucode_getmetadata(uc, &m);
-		if (!sig_count)
-			printf("%03lu: sig 0x%08x, pf mask 0x%02x, "
-				"%04x-%02x-%02x, rev 0x%04x, size %u\n",
-				*counter, cpuid, pf_mask,
-				m.date_year, m.date_month, m.date_day,
-				m.revision, uc_size);
-		else
-			printf("     sig 0x%08x, pf mask 0x%02x, "
-				"%04x-%02x-%02x, rev 0x%04x\n",
-				cpuid, pf_mask,
-				m.date_year, m.date_month, m.date_day,
-				m.revision);
-	}
-
-	return 0;
-}
-
 static int do_process_microcodes(void)
 {
 	intel_ucode_status_t s;
@@ -1324,47 +1598,67 @@ static int do_process_microcodes(void)
 		mcb = mcb->next;
 	}
 
-	print_msg(2, "processed %lu valid microcode(s), %lu signature(s)",
+	print_msg(2, "processed %lu valid microcode(s), %lu signature(s), %lu unique signature(s)",
 		  microcode_interator_data.total_entry_count,
-		  microcode_interator_data.total_signature_count);
+		  microcode_interator_data.total_signature_count,
+		  microcode_interator_data.total_unique_sig_count);
 
 	/*
 	 * we iterate over all selected microcodes only once, either
 	 * to list selected microcodes, or to gather some stats when
-	 * the verbosity is high
+	 * the verbosity is high, or to implement the loose date
+	 * filtering mode.
 	 */
-	if (list_sel_microcodes || verbosity >= 2) {
-		struct intel_uclist_entry *uce;
+	if (list_sel_microcodes || datefilter_loose || verbosity >= 2) {
+		struct intel_uclist_entry *uce = microcodes;
+		struct intel_uclist_entry *ucl = all_microcodes;
 		unsigned long uccount = 0;
 		unsigned long sigcount = 0;
 
 		if (list_sel_microcodes)
 			printf("selected microcodes:\n");
 
-		uce = microcodes;
-		s = INTEL_UCODE_NOERROR;
-		while (uce && s == INTEL_UCODE_NOERROR) {
-			uccount++;
-			s = intel_ucode_foreach_signature(uce->uc,
-					__show_ucode_signature_and_count, &sigcount);
+		while (uce) {
+			/*
+			 * search any extra microcode for loose mode and
+			 * insert it at this point
+			 */
+			if (unlikely(datefilter_loose &&
+			    __datefilter_loose_inplaceinsert(&uce, &ucl)))
+					return 1;
+
+			if (likely(!(uce->flags & INTEL_UCLE_EXTSIG)))
+			    uccount++;
+
+			sigcount++;
+
+			if (list_sel_microcodes) {
+				struct intel_ucode_metadata m;
+
+				intel_ucode_getmetadata(uce->uc, &m);
+				printf("%03lu: sig 0x%08x, pf mask 0x%02x, "
+					"%04x-%02x-%02x, rev 0x%04x, size %u\n",
+					sigcount, uce->cpuid, uce->pf_mask,
+					m.date_year, m.date_month, m.date_day,
+					m.revision, uce->uc_size);
+			}
+
 			uce = uce->next;
 		}
-		if (s == INTEL_UCODE_NOERROR) {
-			print_msg(2, "selected %lu microcode(s), %lu signature(s)",
-				  uccount, sigcount);
-		}
+		print_msg(2, "selected %lu microcode(s), %lu signature(s)",
+			  uccount, sigcount);
 	}
 
 	return 0;
 }
 
-static int do_write_microcode(const char * const filename)
+static int do_write_microcode(const char * const filename, int ft)
 {
 	if (!microcodes)
 		return 0;
 
 	print_msg(1, "Writing selected microcodes to: %s", filename);
-	return write_intel_microcodes(filename, microcodes);
+	return write_intel_microcodes(filename, ft, microcodes);
 }
 
 static int do_upload_microcode(const char * const filename)
@@ -1418,7 +1712,7 @@ static int do_write_named(const char * const dirname)
 
 		memcpy(&e, p, sizeof(e));
 		e.next = NULL;
-		rc = write_intel_microcodes(fn, &e);
+		rc = write_intel_microcodes(fn, 0, &e);
 		if (!rc)
 			count++;
 
@@ -1472,7 +1766,7 @@ static int do_write_firmware(const char * const dirname)
 		/* select all that share the same cpuid */
 		cpuid = p->cpuid;
 		while (p && p->cpuid == cpuid) {
-			add_status = uclist_add_signature(p->id, p->gid,
+			add_status = uclist_merge_signature(p->id, p->gid,
 						p->cpuid, p->pf_mask,
 						p->uc_rev, p->uc, p->uc_size,
 						0, &samecpuid_list);
@@ -1500,7 +1794,7 @@ static int do_write_firmware(const char * const dirname)
 			return EINVAL;
 		}
 
-		rc = write_intel_microcodes(fn, samecpuid_list);
+		rc = write_intel_microcodes(fn, 0, samecpuid_list);
 
 		free_uclist(samecpuid_list);
 		samecpuid_list = NULL;
@@ -1629,7 +1923,7 @@ err_out:
 
 const char program_version[] =
 	PROGNAME " " VERSION "\n"
-	"Copyright (c) 2010-2012 by Henrique de Moraes Holschuh\n\n"
+	"Copyright (c) 2010-2013 by Henrique de Moraes Holschuh\n\n"
 
 	"Based on code from the Linux microcode_intel driver and from\n"
 	"the microcode.ctl package, copyright (c) 2000 by Simon Trimmer\n"
@@ -1657,7 +1951,12 @@ static char cmdline_doc[] =
 
 	"To load microcode data from stdin, use \"-\" as the filename.  "
 	"File type will be assumed to be text (\".dat\"), use option -tb "
-	"to load binary data from stdin.";
+	"to load binary data from stdin.\n\n"
+
+	"To load all files from a directory, specify the directory name.  "
+	"It will not recurse into subdirectories, they will be skipped.\n\n"
+
+	"Empty files and directories will be ignored, and will be skipped.";
 
 enum {
 	IUCODE_ARGP_STRICTCHK = 0x81,
@@ -1670,6 +1969,9 @@ enum {
 	IUCODE_ARGP_NODOWNGRADE,
 	IUCODE_ARGP_DATEBEFORE,
 	IUCODE_ARGP_DATEAFTER,
+	IUCODE_ARGP_DATEFSTRICT,
+	IUCODE_ARGP_DATEFLOOSE,
+	IUCODE_ARGP_EIRFS,
 };
 
 static struct argp_option cmdline_options[] = {
@@ -1684,10 +1986,12 @@ static struct argp_option cmdline_options[] = {
 	   "(type will be selected by filename suffix)",
 	  10 },
 
-	{ NULL, 's', "[!]signature[,pf_mask]", 0,
+	{ NULL, 's', "! | [!]signature[,pf_mask]", 0,
 	   "Select microcodes by the specificed signature and processor "
 	   "flags mask.  Specify more than once to select/unselect more "
-	   "microcodes.  Prefix with ! to unselect microcodes",
+	   "microcodes.  Prefix with ! to unselect microcodes.  Use -s ! "
+	   "to disable the default behaviour of selecting all microcodes "
+	   "when no -s or -S filter is specified",
 	  20 },
 	{ "scan-system", 'S', NULL, 0,
 	   "Select microcodes by scanning all online processors on "
@@ -1712,6 +2016,13 @@ static struct argp_option cmdline_options[] = {
 	{ "date-after", IUCODE_ARGP_DATEAFTER, "YYYY-MM-DD", 0,
 	  "Select only microcodes newer than the specified date",
 	  27 },
+	{ "loose-date-filtering", IUCODE_ARGP_DATEFLOOSE, NULL, 0,
+	  "Consider for selection other revisions (outside of the date range) "
+	  "of every microcode that was selected within the date range",
+	  28 },
+	{ "strict-date-filtering", IUCODE_ARGP_DATEFSTRICT, NULL, 0,
+	  "Select only microcodes strictly within the date range (default)",
+	  28 },
 
 	{ "list",     'l', NULL, 0, "List selected microcode signatures", 30 },
 	{ "list-all", 'L', NULL, 0, "List all microcode signatures",      30 },
@@ -1735,6 +2046,10 @@ static struct argp_option cmdline_options[] = {
 	  "in binary format.  The file name will reflect the microcode "
 	  "signature, mask and revision",
 	  40 },
+	{ "write-earlyfw", IUCODE_ARGP_EIRFS, "file", 0,
+	  "Write selected microcodes to an early initramfs file, which "
+	  "should be prepended to the regular initramfs",
+	  40 },
 
 	{ "overwrite", IUCODE_ARGP_UNLINK, NULL, 0,
 	   "Unlink (remove) destination files before writing",
@@ -1806,7 +2121,7 @@ static void free_filename_list(void)
 	input_files = NULL;
 }
 
-/* -s [!]cpuid[,pf_mask] */
+/* -s ! | [!]cpuid[,pf_mask] */
 static int add_ucode_filter(const char *arg)
 {
 	char *p;
@@ -1821,6 +2136,18 @@ static int add_ucode_filter(const char *arg)
 	if (*arg == '!') {
 		invert = 1;
 		arg++;
+
+		/* handle -s ! */
+		if (isspace(*arg)) {
+			while (isspace(*arg))
+				arg++;
+			if (! *arg)
+			    	return EINVAL;
+		}
+		if (! *arg) {
+			filter_list_allow = 0;
+			return 0;
+		}
 	}
 
 	errno = 0;
@@ -1966,6 +2293,14 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 		write_microcode = strdup(arg);
 		command_line_actions |= IUCODE_DO_WRITEUC;
 		break;
+	case IUCODE_ARGP_EIRFS:
+		if (command_line_actions & IUCODE_DO_WRITEFWE)
+			argp_error(state,
+				   "--write-earlyfw option can be specified only once\n");
+
+		write_early_firmware = strdup(arg);
+		command_line_actions |= IUCODE_DO_WRITEFWE;
+		break;
 	case 'W':
 		if (command_line_actions & IUCODE_DO_WRITEFWN)
 			argp_error(state,
@@ -2008,6 +2343,12 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 			argp_error(state, "invalid date: %s", arg);
 		command_line_actions |= IUCODE_F_UCSELECT;
 		break;
+	case IUCODE_ARGP_DATEFSTRICT:
+		datefilter_loose = 0;
+		break;
+	case IUCODE_ARGP_DATEFLOOSE:
+		datefilter_loose = 1;
+		break;
 
 	case IUCODE_ARGP_STRICTCHK:
 		strict_checks = 1;
@@ -2089,7 +2430,7 @@ int main(int argc, char *argv[])
 			fn = fn->next;
 		}
 
-		if (do_process_microcodes()) {
+		if (microcode_bundles && do_process_microcodes()) {
 			rc = EXIT_SWFAILURE;
 			goto err_exit;
 		}
@@ -2101,7 +2442,9 @@ int main(int argc, char *argv[])
 			if (upload_microcode)
 				rc = do_upload_microcode(upload_microcode);
 			if (!rc && write_microcode)
-				rc = do_write_microcode(write_microcode);
+				rc = do_write_microcode(write_microcode, 0);
+			if (!rc && write_early_firmware)
+				rc = do_write_microcode(write_early_firmware, 1);
 			if (!rc && write_firmware)
 				rc = do_write_firmware(write_firmware);
 			if (!rc && write_named)
@@ -2135,6 +2478,8 @@ err_exit:
 	uc_filter_list = NULL;
 	free_uclist(microcodes);
 	microcodes = NULL;
+	free_uclist(all_microcodes);
+	all_microcodes = NULL;
 	free_intel_microcode_bundles();
 	free_filename_list();
 #endif /* VALGRIND_BUILD */
